#    (c) Copyright 2015 NEC Corporation, All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from unittest import mock

from oslo_utils import uuidutils
from webob import exc

from neutron.tests.unit.api.v2 import test_base as test_api_v2
from neutron_lib.plugins import constants as nconstants

from neutron_vpnaas.extensions import vpn_endpoint_groups
from neutron_vpnaas.extensions import vpnaas
from neutron_vpnaas.services.vpn.common import constants
from neutron_vpnaas.tests import base

_uuid = uuidutils.generate_uuid
_get_path = test_api_v2._get_path


class VpnEndpointGroupsTestPlugin(
        vpnaas.VPNPluginBase,
        vpn_endpoint_groups.VPNEndpointGroupsPluginBase):
    pass


class VpnEndpointGroupsTestCase(base.ExtensionTestCase):

    fmt = 'json'

    def setUp(self):
        super().setUp()
        plural_mappings = {'endpoint_group': 'endpoint-groups'}
        self.setup_extension(
            'neutron_vpnaas.tests.unit.extensions.test_vpn_endpoint_groups.'
            'VpnEndpointGroupsTestPlugin',
            nconstants.VPN,
            vpn_endpoint_groups.Vpn_endpoint_groups,
            'vpn', plural_mappings=plural_mappings,
            use_quota=True)

    def helper_test_endpoint_group_create(self, data):
        """Check that the endpoint_group_create works.

        Uses passed in endpoint group information, which specifies an
        endpoint type and values.
        """
        data['endpoint_group'].update({'tenant_id': _uuid(),
                                       'name': 'my endpoint group',
                                       'description': 'my description'})
        return_value = copy.copy(data['endpoint_group'])
        return_value.update({'id': _uuid()})

        instance = self.plugin.return_value
        instance.create_endpoint_group.return_value = return_value
        res = self.api.post(_get_path('vpn/endpoint-groups', fmt=self.fmt),
                            self.serialize(data),
                            content_type='application/%s' % self.fmt)
        self.assertEqual(1, instance.create_endpoint_group.call_count)
        self.assertEqual(exc.HTTPCreated.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('endpoint_group', res)
        self.assertDictSupersetOf(return_value, res['endpoint_group'])

    def test_create_cidr_endpoint_group_create(self):
        """Test creation of CIDR type endpoint group."""
        data = {'endpoint_group':
                {'type': constants.CIDR_ENDPOINT,
                 'endpoints': ['10.10.10.0/24', '20.20.20.0/24']}}
        self.helper_test_endpoint_group_create(data)

    def test_create_subnet_endpoint_group_create(self):
        """Test creation of subnet type endpoint group."""
        data = {'endpoint_group':
                {'type': constants.SUBNET_ENDPOINT,
                 'endpoints': [_uuid(), _uuid()]}}
        self.helper_test_endpoint_group_create(data)

    def test_create_vlan_endpoint_group_create(self):
        """Test creation of VLAN type endpoint group."""
        data = {'endpoint_group':
                {'type': constants.VLAN_ENDPOINT,
                 'endpoints': ['100', '200', '300', '400']}}
        self.helper_test_endpoint_group_create(data)

    def test_get_endpoint_group(self):
        """Test show for endpoint group."""
        endpoint_group_id = _uuid()
        return_value = {'id': endpoint_group_id,
                        'tenant_id': _uuid(),
                        'name': 'my-endpoint-group',
                        'description': 'my endpoint group',
                        'type': constants.CIDR_ENDPOINT,
                        'endpoints': ['10.10.10.0/24']}

        instance = self.plugin.return_value
        instance.get_endpoint_group.return_value = return_value

        res = self.api.get(_get_path('vpn/endpoint-groups',
                                     id=endpoint_group_id,
                                     fmt=self.fmt))

        instance.get_endpoint_group.assert_called_with(mock.ANY,
                                                       endpoint_group_id,
                                                       fields=mock.ANY)
        self.assertEqual(res.status_int, exc.HTTPOk.code)
        res = self.deserialize(res)
        self.assertIn('endpoint_group', res)
        self.assertEqual(res['endpoint_group'], return_value)

    def test_endpoint_group_list(self):
        """Test listing all endpoint groups."""
        return_value = [{'id': _uuid(),
                         'tenant_id': _uuid(),
                         'name': 'my-endpoint-group',
                         'description': 'my endpoint group',
                         'type': constants.CIDR_ENDPOINT,
                         'endpoints': ['10.10.10.0/24']},
                        {'id': _uuid(),
                         'tenant_id': _uuid(),
                         'name': 'another-endpoint-group',
                         'description': 'second endpoint group',
                         'type': constants.VLAN_ENDPOINT,
                         'endpoints': ['100', '200', '300']}]

        instance = self.plugin.return_value
        instance.get_endpoint_groups.return_value = return_value

        res = self.api.get(_get_path('vpn/endpoint-groups', fmt=self.fmt))

        instance.get_endpoint_groups.assert_called_with(mock.ANY,
                                                        fields=mock.ANY,
                                                        filters=mock.ANY)
        self.assertEqual(res.status_int, exc.HTTPOk.code)

    def test_endpoint_group_delete(self):
        """Test deleting an endpoint group."""
        self._test_entity_delete('endpoint_group')

    def test_endpoint_group_update(self):
        """Test updating endpoint_group."""
        endpoint_group_id = _uuid()
        update_data = {'endpoint_group': {'description': 'new description'}}
        return_value = {'id': endpoint_group_id,
                        'tenant_id': _uuid(),
                        'name': 'my-endpoint-group',
                        'description': 'new_description',
                        'type': constants.CIDR_ENDPOINT,
                        'endpoints': ['10.10.10.0/24']}

        instance = self.plugin.return_value
        instance.update_endpoint_group.return_value = return_value

        res = self.api.put(_get_path('vpn/endpoint-groups',
                                     id=endpoint_group_id,
                                     fmt=self.fmt),
                           self.serialize(update_data))

        instance.update_endpoint_group.assert_called_with(
            mock.ANY, endpoint_group_id, endpoint_group=update_data)
        self.assertEqual(res.status_int, exc.HTTPOk.code)
        res = self.deserialize(res)
        self.assertIn('endpoint_group', res)
        self.assertEqual(res['endpoint_group'], return_value)

    def test_fail_updating_endpoints_in_endpoint_group(self):
        """Test fails to update the endpoints in an endpoint group.

        This documents that we are not allowing endpoints to be updated
        (currently), as doing so, implies that the connection using the
        enclosing endpoint group would also need to be updated. For now,
        a new endpoint group can be created, and the connection can be
        updated to point to the new endpoint group.
        """
        endpoint_group_id = _uuid()
        update_data = {'endpoint_group': {'endpoints': ['10.10.10.0/24']}}
        res = self.api.put(_get_path('vpn/endpoint-groups',
                                     id=endpoint_group_id,
                                     fmt=self.fmt),
                           params=self.serialize(update_data),
                           expect_errors=True)
        self.assertEqual(exc.HTTPBadRequest.code, res.status_int)



# Copyright 2015, Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import socket
from unittest import mock

from neutron.db import l3_db
from neutron_lib import context as n_ctx
from neutron_lib import exceptions as nexception
from neutron_lib.exceptions import vpn as vpn_exception
from neutron_lib.plugins import constants as nconstants
from neutron_lib.plugins import directory
from oslo_utils import uuidutils
from sqlalchemy.orm import query

from neutron_vpnaas.db.vpn import vpn_validator
from neutron_vpnaas.services.vpn.common import constants as v_constants
from neutron_vpnaas.tests import base

_uuid = uuidutils.generate_uuid

FAKE_ROUTER_ID = _uuid()
FAKE_ROUTER = {l3_db.EXTERNAL_GW_INFO: FAKE_ROUTER_ID}
FAKE_SUBNET_ID = _uuid()
IPV4 = 4
IPV6 = 6


class TestVpnValidation(base.BaseTestCase):

    def setUp(self):
        super().setUp()
        self.l3_plugin = mock.Mock()
        self.core_plugin = mock.Mock()
        directory.add_plugin(nconstants.CORE, self.core_plugin)
        directory.add_plugin(nconstants.L3, self.l3_plugin)
        self.context = n_ctx.Context('some_user', 'some_tenant')
        self.validator = vpn_validator.VpnReferenceValidator()
        self.router = mock.Mock()
        self.router.gw_port = {'fixed_ips': [{'ip_address': '10.0.0.99'}]}

    def test_non_public_router_for_vpn_service(self):
        """Failure test of service validate, when router missing ext. I/F."""
        self.l3_plugin.get_router.return_value = {}  # No external gateway
        vpnservice = {'router_id': 123, 'subnet_id': 456}
        self.assertRaises(vpn_exception.RouterIsNotExternal,
                          self.validator.validate_vpnservice,
                          self.context, vpnservice)

    def test_subnet_not_connected_for_vpn_service(self):
        """Failure test of service validate, when subnet not on router."""
        self.l3_plugin.get_router.return_value = FAKE_ROUTER
        self.core_plugin.get_ports.return_value = None
        vpnservice = {'router_id': FAKE_ROUTER_ID, 'subnet_id': FAKE_SUBNET_ID}
        self.assertRaises(vpn_exception.SubnetIsNotConnectedToRouter,
                          self.validator.validate_vpnservice,
                          self.context, vpnservice)

    def test_defaults_for_ipsec_site_connections_on_create(self):
        """Check that defaults are applied correctly.

        MTU has a default and will always be present on create.
        However, the DPD settings do not have a default, so
        database create method will assign default values for any
        missing. In addition, the DPD dict will be flattened
        for storage into the database, so we'll do it as part of
        assigning defaults.
        """
        ipsec_sitecon = {}
        self.validator.assign_sensible_ipsec_sitecon_defaults(ipsec_sitecon)
        expected = {
            'dpd_action': 'hold',
            'dpd_timeout': 120,
            'dpd_interval': 30
        }
        self.assertEqual(expected, ipsec_sitecon)

        ipsec_sitecon = {'dpd': {'interval': 50}}
        self.validator.assign_sensible_ipsec_sitecon_defaults(ipsec_sitecon)
        expected = {
            'dpd': {'interval': 50},
            'dpd_action': 'hold',
            'dpd_timeout': 120,
            'dpd_interval': 50
        }
        self.assertEqual(expected, ipsec_sitecon)

    def test_resolve_peer_address_with_ipaddress(self):
        ipsec_sitecon = {'peer_address': '10.0.0.9'}
        self.validator._validate_peer_address = mock.Mock()
        self.validator.resolve_peer_address(ipsec_sitecon, self.router)
        self.assertEqual('10.0.0.9', ipsec_sitecon['peer_address'])
        self.validator._validate_peer_address.assert_called_once_with(
            IPV4, self.router)

    def test_resolve_peer_address_with_fqdn(self):
        with mock.patch.object(socket, 'getaddrinfo') as mock_getaddr_info:
            mock_getaddr_info.return_value = [(2, 1, 6, '',
                                              ('10.0.0.9', 0))]
            ipsec_sitecon = {'peer_address': 'fqdn.peer.addr'}
            self.validator._validate_peer_address = mock.Mock()
            self.validator.resolve_peer_address(ipsec_sitecon, self.router)
            self.assertEqual('10.0.0.9', ipsec_sitecon['peer_address'])
            self.validator._validate_peer_address.assert_called_once_with(
                IPV4, self.router)

    def test_resolve_peer_address_with_invalid_fqdn(self):
        with mock.patch.object(socket, 'getaddrinfo') as mock_getaddr_info:
            def getaddr_info_failer(*args, **kwargs):
                raise socket.gaierror()
            mock_getaddr_info.side_effect = getaddr_info_failer
            ipsec_sitecon = {'peer_address': 'fqdn.invalid'}
            self.assertRaises(vpn_exception.VPNPeerAddressNotResolved,
                              self.validator.resolve_peer_address,
                              ipsec_sitecon, self.router)

    def helper_validate_peer_address(self, fixed_ips, ip_version,
                                     expected_exception=False):
        self.router.id = FAKE_ROUTER_ID
        self.router.gw_port = {'fixed_ips': fixed_ips}
        try:
            self.validator._validate_peer_address(ip_version, self.router)
            if expected_exception:
                self.fail("No exception raised for invalid peer address")
        except vpn_exception.ExternalNetworkHasNoSubnet:
            if not expected_exception:
                self.fail("exception for valid peer address raised")

    def test_validate_peer_address(self):
        # validate ipv4 peer_address with ipv4 gateway
        fixed_ips = [{'ip_address': '10.0.0.99'}]
        self.helper_validate_peer_address(fixed_ips, IPV4)

        # validate ipv6 peer_address with ipv6 gateway
        fixed_ips = [{'ip_address': '2001::1'}]
        self.helper_validate_peer_address(fixed_ips, IPV6)

        # validate ipv6 peer_address with both ipv4 and ipv6 gateways
        fixed_ips = [{'ip_address': '2001::1'}, {'ip_address': '10.0.0.99'}]
        self.helper_validate_peer_address(fixed_ips, IPV6)

        # validate ipv4 peer_address with both ipv4 and ipv6 gateways
        fixed_ips = [{'ip_address': '2001::1'}, {'ip_address': '10.0.0.99'}]
        self.helper_validate_peer_address(fixed_ips, IPV4)

        # validate ipv4 peer_address with ipv6 gateway
        fixed_ips = [{'ip_address': '2001::1'}]
        self.helper_validate_peer_address(fixed_ips, IPV4,
                                          expected_exception=True)

        # validate ipv6 peer_address with ipv4 gateway
        fixed_ips = [{'ip_address': '10.0.0.99'}]
        self.helper_validate_peer_address(fixed_ips, IPV6,
                                          expected_exception=True)

    def test_defaults_for_ipsec_site_connections_on_update(self):
        """Check that defaults are used for any values not specified."""
        ipsec_sitecon = {}
        prev_connection = {'peer_cidrs': [{'cidr': '10.0.0.0/24'},
                                          {'cidr': '20.0.0.0/24'}],
                           'local_ep_group_id': None,
                           'peer_ep_group_id': None,
                           'dpd_action': 'clear',
                           'dpd_timeout': 500,
                           'dpd_interval': 250}
        self.validator.assign_sensible_ipsec_sitecon_defaults(ipsec_sitecon,
                                                              prev_connection)
        expected = {
            'peer_cidrs': ['10.0.0.0/24', '20.0.0.0/24'],
            'local_ep_group_id': None,
            'peer_ep_group_id': None,
            'dpd_action': 'clear',
            'dpd_timeout': 500,
            'dpd_interval': 250
        }
        self.assertEqual(expected, ipsec_sitecon)

        ipsec_sitecon = {'dpd': {'timeout': 200}}
        local_epg_id = _uuid()
        peer_epg_id = _uuid()
        prev_connection = {'peer_cidrs': [],
                           'local_ep_group_id': local_epg_id,
                           'peer_ep_group_id': peer_epg_id,
                           'dpd_action': 'clear',
                           'dpd_timeout': 500,
                           'dpd_interval': 100}
        self.validator.assign_sensible_ipsec_sitecon_defaults(ipsec_sitecon,
                                                              prev_connection)
        expected = {
            'peer_cidrs': [],
            'local_ep_group_id': local_epg_id,
            'peer_ep_group_id': peer_epg_id,
            'dpd': {'timeout': 200},
            'dpd_action': 'clear',
            'dpd_timeout': 200,
            'dpd_interval': 100
        }
        self.assertEqual(expected, ipsec_sitecon)

    def test_bad_dpd_settings_on_create(self):
        """Failure tests of DPD settings for IPSec conn during create."""
        ipsec_sitecon = {'dpd_action': 'hold', 'dpd_interval': 100,
                         'dpd_timeout': 100}
        self.assertRaises(
            vpn_exception.IPsecSiteConnectionDpdIntervalValueError,
            self.validator._check_dpd, ipsec_sitecon)
        ipsec_sitecon = {'dpd_action': 'hold', 'dpd_interval': 100,
                         'dpd_timeout': 99}
        self.assertRaises(
            vpn_exception.IPsecSiteConnectionDpdIntervalValueError,
            self.validator._check_dpd, ipsec_sitecon)

    def test_bad_mtu_for_ipsec_connection(self):
        """Failure test of invalid MTU values for IPSec conn create/update."""
        ip_version_limits = self.validator.IP_MIN_MTU
        for version, limit in ip_version_limits.items():
            ipsec_sitecon = {'mtu': limit - 1}
            self.assertRaises(
                vpn_exception.IPsecSiteConnectionMtuError,
                self.validator._check_mtu,
                self.context, ipsec_sitecon['mtu'], version)

    def test_endpoints_all_cidrs_in_endpoint_group(self):
        """All endpoints in the endpoint group are valid CIDRs."""
        endpoint_group = {'type': v_constants.CIDR_ENDPOINT,
                          'endpoints': ['10.10.10.0/24', '20.20.20.0/24']}
        try:
            self.validator.validate_endpoint_group(self.context,
                                                   endpoint_group)
        except Exception:
            self.fail("All CIDRs in endpoint_group should be valid")

    def test_endpoints_all_subnets_in_endpoint_group(self):
        """All endpoints in the endpoint group are valid subnets."""
        endpoint_group = {'type': v_constants.SUBNET_ENDPOINT,
                          'endpoints': [_uuid(), _uuid()]}
        try:
            self.validator.validate_endpoint_group(self.context,
                                                   endpoint_group)
        except Exception:
            self.fail("All subnets in endpoint_group should be valid")

    def test_mixed_endpoint_types_in_endpoint_group(self):
        """Fail when mixing types of endpoints in endpoint group."""
        endpoint_group = {'type': v_constants.CIDR_ENDPOINT,
                          'endpoints': ['10.10.10.0/24', _uuid()]}
        self.assertRaises(vpn_exception.InvalidEndpointInEndpointGroup,
                          self.validator.validate_endpoint_group,
                          self.context, endpoint_group)
        endpoint_group = {'type': v_constants.SUBNET_ENDPOINT,
                          'endpoints': [_uuid(), '10.10.10.0/24']}
        self.assertRaises(vpn_exception.InvalidEndpointInEndpointGroup,
                          self.validator.validate_endpoint_group,
                          self.context, endpoint_group)

    def test_missing_endpoints_for_endpoint_group(self):
        endpoint_group = {'type': v_constants.CIDR_ENDPOINT,
                          'endpoints': []}
        self.assertRaises(vpn_exception.MissingEndpointForEndpointGroup,
                          self.validator.validate_endpoint_group,
                          self.context, endpoint_group)

    def test_fail_bad_cidr_in_endpoint_group(self):
        """Testing catches bad CIDR.

        Just check one case, as CIDR validator used has good test coverage.
        """
        endpoint_group = {'type': v_constants.CIDR_ENDPOINT,
                          'endpoints': ['10.10.10.10/24', '20.20.20.1']}
        self.assertRaises(vpn_exception.InvalidEndpointInEndpointGroup,
                          self.validator.validate_endpoint_group,
                          self.context, endpoint_group)

    def test_unknown_subnet_in_endpoint_group(self):
        subnet_id = _uuid()
        self.core_plugin.get_subnet.side_effect = nexception.SubnetNotFound(
            subnet_id=subnet_id)
        endpoint_group = {'type': v_constants.SUBNET_ENDPOINT,
                          'endpoints': [subnet_id]}
        self.assertRaises(vpn_exception.NonExistingSubnetInEndpointGroup,
                          self.validator.validate_endpoint_group,
                          self.context, endpoint_group)

    def test_fail_subnets_not_on_same_router_for_endpoint_group(self):
        """Detect when local endpoints not on the same router."""
        subnet1 = {'id': _uuid(), 'ip_version': 4}
        subnet2 = {'id': _uuid(), 'ip_version': 4}
        router = _uuid()
        multiple_subnets = [subnet1, subnet2]
        port_mock = mock.patch.object(self.core_plugin, "get_ports").start()
        port_mock.side_effect = ['dummy info', None]
        self.assertRaises(vpn_exception.SubnetIsNotConnectedToRouter,
                          self.validator._check_local_subnets_on_router,
                          self.context, router, multiple_subnets)

    def test_ipsec_conn_local_endpoints_same_ip_version(self):
        """Check local endpoint subnets all have same IP version."""
        endpoint_group_id = _uuid()
        subnet1 = {'ip_version': 4}
        subnet2 = {'ip_version': 4}
        single_subnet = [subnet1]
        version = self.validator._check_local_endpoint_ip_versions(
            endpoint_group_id, single_subnet)
        self.assertEqual(4, version)
        multiple_subnets = [subnet1, subnet2]
        version = self.validator._check_local_endpoint_ip_versions(
            endpoint_group_id, multiple_subnets)
        self.assertEqual(4, version)

    def test_fail_ipsec_conn_local_endpoints_mixed_ip_version(self):
        """Fail when mixed IP versions in local endpoints."""
        endpoint_group_id = _uuid()
        subnet1 = {'ip_version': 6}
        subnet2 = {'ip_version': 4}
        mixed_subnets = [subnet1, subnet2]
        self.assertRaises(vpn_exception.MixedIPVersionsForIPSecEndpoints,
                          self.validator._check_local_endpoint_ip_versions,
                          endpoint_group_id, mixed_subnets)

    def test_ipsec_conn_peer_endpoints_same_ip_version(self):
        """Check all CIDRs have the same IP version."""
        endpoint_group_id = _uuid()
        one_cidr = ['2002:0a00::/48']
        version = self.validator._check_peer_endpoint_ip_versions(
            endpoint_group_id, one_cidr)
        self.assertEqual(6, version)
        multiple_cidr = ['10.10.10.0/24', '20.20.20.0/24']
        version = self.validator._check_peer_endpoint_ip_versions(
            endpoint_group_id, multiple_cidr)
        self.assertEqual(4, version)

    def test_fail_ipsec_conn_peer_endpoints_mixed_ip_version(self):
        """Fail when mixed IP versions in peer endpoints."""
        endpoint_group_id = _uuid()
        mixed_cidrs = ['10.10.10.0/24', '2002:1400::/48']
        self.assertRaises(vpn_exception.MixedIPVersionsForIPSecEndpoints,
                          self.validator._check_peer_endpoint_ip_versions,
                          endpoint_group_id, mixed_cidrs)

    def test_fail_ipsec_conn_locals_and_peers_different_ip_version(self):
        """Ensure catch when local and peer IP versions are not the same."""
        self.assertRaises(vpn_exception.MixedIPVersionsForIPSecConnection,
                          self.validator._validate_compatible_ip_versions,
                          4, 6)

    def test_fail_ipsec_conn_no_subnet_requiring_endpoint_groups(self):
        """When no subnet, connection must use endpoints.

        This means both endpoint groups must be present, and peer_cidrs
        cannot be used.
        """
        subnet = None
        ipsec_sitecon = {'peer_cidrs': ['10.0.0.0/24'],
                         'local_ep_group_id': 'local-epg-id',
                         'peer_ep_group_id': 'peer-epg-id'}
        self.assertRaises(vpn_exception.PeerCidrsInvalid,
                          self.validator.validate_ipsec_conn_optional_args,
                          ipsec_sitecon, subnet)

        ipsec_sitecon = {'peer_cidrs': [],
                         'local_ep_group_id': None,
                         'peer_ep_group_id': 'peer-epg-id'}
        self.assertRaises(vpn_exception.MissingRequiredEndpointGroup,
                          self.validator.validate_ipsec_conn_optional_args,
                          ipsec_sitecon, subnet)

        ipsec_sitecon = {'peer_cidrs': [],
                         'local_ep_group_id': 'local-epg-id',
                         'peer_ep_group_id': None}
        self.assertRaises(vpn_exception.MissingRequiredEndpointGroup,
                          self.validator.validate_ipsec_conn_optional_args,
                          ipsec_sitecon, subnet)

        ipsec_sitecon = {'peer_cidrs': [],
                         'local_ep_group_id': None,
                         'peer_ep_group_id': None}
        self.assertRaises(vpn_exception.MissingRequiredEndpointGroup,
                          self.validator.validate_ipsec_conn_optional_args,
                          ipsec_sitecon, subnet)

    def test_fail_ipsec_conn_subnet_requiring_peer_cidrs(self):
        """When legacy mode, no endpoint groups.

        This means neither endpoint group can be specified, and the peer_cidrs
        must be present.
        """
        subnet = {'id': FAKE_SUBNET_ID}
        ipsec_sitecon = {'peer_cidrs': [],
                         'local_ep_group_id': None,
                         'peer_ep_group_id': None}
        self.assertRaises(vpn_exception.MissingPeerCidrs,
                          self.validator.validate_ipsec_conn_optional_args,
                          ipsec_sitecon, subnet)

        ipsec_sitecon = {'peer_cidrs': ['10.0.0.0/24'],
                         'local_ep_group_id': 'local-epg-id',
                         'peer_ep_group_id': None}
        self.assertRaises(vpn_exception.InvalidEndpointGroup,
                          self.validator.validate_ipsec_conn_optional_args,
                          ipsec_sitecon, subnet)

        ipsec_sitecon = {'peer_cidrs': ['10.0.0.0/24'],
                         'local_ep_group_id': None,
                         'peer_ep_group_id': 'peer-epg-id'}
        self.assertRaises(vpn_exception.InvalidEndpointGroup,
                          self.validator.validate_ipsec_conn_optional_args,
                          ipsec_sitecon, subnet)

        ipsec_sitecon = {'peer_cidrs': ['10.0.0.0/24'],
                         'local_ep_group_id': 'local-epg-id',
                         'peer_ep_group_id': 'peer-epg-id'}
        self.assertRaises(vpn_exception.InvalidEndpointGroup,
                          self.validator.validate_ipsec_conn_optional_args,
                          ipsec_sitecon, subnet)

    def test_ipsec_conn_get_local_subnets(self):
        subnet1 = _uuid()
        subnet2 = _uuid()
        expected_subnets = [subnet1, subnet2]
        local_epg = {'id': _uuid(),
                     'type': v_constants.SUBNET_ENDPOINT,
                     'endpoints': expected_subnets}
        query_mock = mock.patch.object(query.Query, 'all').start()
        query_mock.return_value = expected_subnets
        subnets = self.validator._get_local_subnets(self.context, local_epg)
        self.assertEqual(expected_subnets, subnets)

    def test_ipsec_conn_get_peer_cidrs(self):
        expected_cidrs = ['10.10.10.10/24', '20.20.20.20/24']
        peer_epg = {'id': 'should-be-cidrs',
                    'type': v_constants.CIDR_ENDPOINT,
                    'endpoints': expected_cidrs}
        cidrs = self.validator._get_peer_cidrs(peer_epg)
        self.assertEqual(expected_cidrs, cidrs)

    def test_ipsec_conn_check_peer_cidrs(self):
        peer_cidrs = ['10.10.10.0/24', '20.20.20.0/24']
        try:
            self.validator._check_peer_cidrs(peer_cidrs)
        except Exception:
            self.fail("All peer cidrs format should be valid")

    def test_fail_ipsec_conn_peer_cidrs_with_invalid_format(self):
        peer_cidrs = ['invalid_cidr']
        self.assertRaises(vpn_exception.IPsecSiteConnectionPeerCidrError,
                          self.validator._check_peer_cidrs,
                          peer_cidrs)
        peer_cidrs = ['192/24']
        self.assertRaises(vpn_exception.IPsecSiteConnectionPeerCidrError,
                          self.validator._check_peer_cidrs,
                          peer_cidrs)

    def test_fail_ipsec_conn_endpoint_group_types(self):
        local_epg = {'id': 'should-be-subnets',
                     'type': v_constants.CIDR_ENDPOINT,
                     'endpoints': ['10.10.10.10/24', '20.20.20.20/24']}
        self.assertRaises(vpn_exception.WrongEndpointGroupType,
                          self.validator._get_local_subnets,
                          self.context, local_epg)
        peer_epg = {'id': 'should-be-cidrs',
                    'type': v_constants.SUBNET_ENDPOINT,
                    'endpoints': [_uuid(), _uuid()]}
        self.assertRaises(vpn_exception.WrongEndpointGroupType,
                          self.validator._get_peer_cidrs,
                          peer_epg)

    def test_validate_ipsec_conn_for_endpoints(self):
        """Check upper-level validation method for endpoint groups.

        Tests the happy path for doing general validation of the IPSec
        connection, calling all the sub-checks for an endpoint group case.
        """
        subnet1 = {'id': _uuid(), 'ip_version': 4}
        subnet2 = {'id': _uuid(), 'ip_version': 4}
        local_subnets = [subnet1, subnet2]
        local_epg_id = _uuid()
        local_epg = {'id': local_epg_id,
                     'type': v_constants.SUBNET_ENDPOINT,
                     'endpoints': local_subnets}
        # Mock getting the subnets from the IDs
        query_mock = mock.patch.object(query.Query, 'all').start()
        query_mock.return_value = local_subnets
        # Mock that subnet is on router
        port_mock = mock.patch.object(self.core_plugin, "get_ports").start()
        port_mock.side_effect = ['dummy info', 'more dummy info']

        peer_epg_id = _uuid()
        peer_cidrs = ['10.10.10.10/24', '20.20.20.20/24']
        peer_epg = {'id': peer_epg_id,
                    'type': v_constants.CIDR_ENDPOINT,
                    'endpoints': peer_cidrs}

        ipsec_sitecon = {'local_ep_group_id': local_epg_id,
                         'local_epg_subnets': local_epg,
                         'peer_ep_group_id': peer_epg_id,
                         'peer_epg_cidrs': peer_epg,
                         'mtu': 2000,
                         'dpd_action': 'hold',
                         'dpd_interval': 30,
                         'dpd_timeout': 120}
        local_version = None
        vpnservice = {'router_id': _uuid()}
        self.validator.validate_ipsec_site_connection(
            self.context, ipsec_sitecon, local_version, vpnservice)

    # NOTE: Following are tests for the older API, providing some additional
    # coverage.

    def test_ipsec_conn_peer_cidrs_same_ip_version(self):
        """Check legacy peer_cidrs have same IP version."""
        one_cidr = ['2002:0a00::/48']
        version = self.validator._check_peer_cidrs_ip_versions(one_cidr)
        self.assertEqual(6, version)
        multiple_cidrs = ['10.10.10.0/24', '20.20.20.0/24']
        version = self.validator._check_peer_cidrs_ip_versions(multiple_cidrs)
        self.assertEqual(4, version)

    def test_fail_ipsec_conn_peer_cidrs_mixed_ip_version(self):
        mixed_cidrs = ['2002:0a00::/48', '20.20.20.0/24']
        self.assertRaises(vpn_exception.MixedIPVersionsForPeerCidrs,
                          self.validator._check_peer_cidrs_ip_versions,
                          mixed_cidrs)



# Copyright 2016 <Yi Jing Zhu/IBM>
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""support sha256

Revision ID: fe637dc3f042
Revises: 28ee739a7e4b
Create Date: 2016-04-08 22:33:53.286083

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'fe637dc3f042'
down_revision = '28ee739a7e4b'

new_auth = sa.Enum('sha1', 'sha256', name='vpn_auth_algorithms')


def upgrade():
    op.alter_column('ikepolicies', 'auth_algorithm', type_=new_auth,
                    nullable=False, do_drop=False)
    op.alter_column('ipsecpolicies', 'auth_algorithm', type_=new_auth,
                    nullable=False, do_rename=False, do_create=False)



#    (c) Copyright 2013 Hewlett-Packard Development Company, L.P.
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.db import servicetype_db as st_db
from neutron.services import provider_configuration as pconf
from neutron.services import service_base
from neutron_lib import context as ncontext
from neutron_lib import exceptions as lib_exc
from neutron_lib.exceptions import flavors as flav_exc
from neutron_lib.plugins import constants
from neutron_lib.plugins import directory
from oslo_log import log as logging

from neutron_vpnaas.db.vpn import vpn_db
from neutron_vpnaas.extensions import vpn_flavors

LOG = logging.getLogger(__name__)


def add_provider_configuration(type_manager, service_type):
    type_manager.add_provider_configuration(
        service_type,
        pconf.ProviderConfiguration('neutron_vpnaas', service_type))


class VPNPlugin(vpn_db.VPNPluginDb):

    """Implementation of the VPN Service Plugin.

    This class manages the workflow of VPNaaS request/response.
    Most DB related works are implemented in class
    vpn_db.VPNPluginDb.
    """
    supported_extension_aliases = ["vpnaas",
                                   "vpn-endpoint-groups",
                                   "service-type",
                                   "vpn-flavors",
                                   "vpn-aes-ctr"]
    path_prefix = "/vpn"


class VPNDriverPlugin(VPNPlugin, vpn_db.VPNPluginRpcDbMixin):
    """VpnPlugin which supports VPN Service Drivers."""
    #TODO(nati) handle ikepolicy and ipsecpolicy update usecase
    def __init__(self):
        super().__init__()
        self.service_type_manager = st_db.ServiceTypeManager.get_instance()
        add_provider_configuration(self.service_type_manager, constants.VPN)
        # Load the service driver from neutron.conf.
        self.drivers, self.default_provider = service_base.load_drivers(
            constants.VPN, self)
        self._check_orphan_vpnservice_associations()
        # Associate driver names to driver objects
        for driver_name, driver in self.drivers.items():
            driver.name = driver_name
        LOG.info(("VPN plugin using service drivers: %(service_drivers)s, "
                  "default: %(default_driver)s"),
                 {'service_drivers': self.drivers.keys(),
                  'default_driver': self.default_provider})
        vpn_db.subscribe()

    @property
    def _flavors_plugin(self):
        return directory.get_plugin(constants.FLAVORS)

    def start_rpc_listeners(self):
        servers = []
        for driver_name, driver in self.drivers.items():
            if hasattr(driver, 'start_rpc_listeners'):
                servers.extend(driver.start_rpc_listeners())
        return servers

    def _check_orphan_vpnservice_associations(self):
        context = ncontext.get_admin_context()
        vpnservices = self.get_vpnservices(context)
        vpnservice_ids = [vpnservice['id'] for vpnservice in vpnservices]

        stm = self.service_type_manager
        provider_names = stm.get_provider_names_by_resource_ids(
            context, vpnservice_ids)

        lost_providers = set()
        lost_vpnservices = []
        for vpnservice_id, provider in provider_names.items():
            if provider not in self.drivers:
                lost_providers.add(provider)
                lost_vpnservices.append(vpnservice_id)
        if lost_providers or lost_vpnservices:
            # Provider are kept internally, we need to inform users about
            # the related VPN services.
            msg = (
                "Delete associated vpnservices %(vpnservices)s before "
                "removing providers %(providers)s."
            ) % {'vpnservices': lost_vpnservices,
                 'providers': list(lost_providers)}
            LOG.exception(msg)
            raise SystemExit(msg)

        # Deal with upgrade. Associate existing VPN services to default
        # provider.
        unasso_vpnservices = [
            vpnservice_id for vpnservice_id in vpnservice_ids
            if vpnservice_id not in provider_names]
        if unasso_vpnservices:
            LOG.info(
                ("Associating VPN services %(unasso_vpnservices)s to "
                 "default provider %(default_provider)s."),
                {'unasso_vpnservices': unasso_vpnservices,
                 'default_provider': self.default_provider})
            for vpnservice_id in unasso_vpnservices:
                stm.add_resource_association(
                    context, constants.VPN,
                    self.default_provider, vpnservice_id)

    def _get_provider_for_flavor(self, context, flavor_id):
        if flavor_id:
            if self._flavors_plugin is None:
                raise vpn_flavors.FlavorsPluginNotLoaded()

            fl_db = self._flavors_plugin.get_flavor(context, flavor_id)
            if fl_db['service_type'] != constants.VPN:
                raise lib_exc.InvalidServiceType(
                    service_type=fl_db['service_type'])
            if not fl_db['enabled']:
                raise flav_exc.FlavorDisabled()
            providers = self._flavors_plugin.get_flavor_next_provider(
                context, fl_db['id'])
            provider = providers[0].get('provider')
            if provider not in self.drivers:
                raise vpn_flavors.NoProviderFoundForFlavor(flavor_id=flavor_id)
        else:
            # Use default provider
            provider = self.default_provider

        LOG.debug("Selected provider %s", provider)
        return provider

    def _get_driver_for_vpnservice(self, context, vpnservice):
        stm = self.service_type_manager
        provider_names = stm.get_provider_names_by_resource_ids(
            context, [vpnservice['id']])
        provider = provider_names.get(vpnservice['id'])
        return self.drivers[provider]

    def _get_driver_for_ipsec_site_connection(self, context,
                                              ipsec_site_connection):
        # Only vpnservice_id is required as the vpnservice should be already
        # associated with a provider after its creation.
        vpnservice = {'id': ipsec_site_connection['vpnservice_id']}
        return self._get_driver_for_vpnservice(context, vpnservice)

    def create_ipsec_site_connection(self, context, ipsec_site_connection):
        driver = self._get_driver_for_ipsec_site_connection(
            context, ipsec_site_connection['ipsec_site_connection'])
        driver.validator.validate_ipsec_site_connection(
            context,
            ipsec_site_connection['ipsec_site_connection'])
        ipsec_site_connection = super().create_ipsec_site_connection(
                context, ipsec_site_connection)
        driver.create_ipsec_site_connection(context, ipsec_site_connection)
        return ipsec_site_connection

    def delete_ipsec_site_connection(self, context, ipsec_conn_id):
        ipsec_site_connection = self.get_ipsec_site_connection(
            context, ipsec_conn_id)
        super().delete_ipsec_site_connection(
            context, ipsec_conn_id)
        driver = self._get_driver_for_ipsec_site_connection(
            context, ipsec_site_connection)
        driver.delete_ipsec_site_connection(context, ipsec_site_connection)

    def update_ipsec_site_connection(
            self, context,
            ipsec_conn_id, ipsec_site_connection):
        old_ipsec_site_connection = self.get_ipsec_site_connection(
            context, ipsec_conn_id)
        driver = self._get_driver_for_ipsec_site_connection(
            context, old_ipsec_site_connection)
        driver.validator.validate_ipsec_site_connection(
            context,
            ipsec_site_connection['ipsec_site_connection'])
        ipsec_site_connection = super().update_ipsec_site_connection(
                context,
                ipsec_conn_id,
                ipsec_site_connection)
        driver.update_ipsec_site_connection(
            context, old_ipsec_site_connection, ipsec_site_connection)
        return ipsec_site_connection

    def create_vpnservice(self, context, vpnservice):
        provider = self._get_provider_for_flavor(
            context, vpnservice['vpnservice'].get('flavor_id'))
        vpnservice = super().create_vpnservice(context, vpnservice)
        self.service_type_manager.add_resource_association(
            context, constants.VPN, provider, vpnservice['id'])
        driver = self.drivers[provider]
        driver.create_vpnservice(context, vpnservice)
        return vpnservice

    def update_vpnservice(self, context, vpnservice_id, vpnservice):
        old_vpn_service = self.get_vpnservice(context, vpnservice_id)
        new_vpn_service = super().update_vpnservice(context, vpnservice_id,
                                                    vpnservice)
        driver = self._get_driver_for_vpnservice(context, old_vpn_service)
        driver.update_vpnservice(context, old_vpn_service, new_vpn_service)
        return new_vpn_service

    def delete_vpnservice(self, context, vpnservice_id):
        vpnservice = self._get_vpnservice(context, vpnservice_id)
        super().delete_vpnservice(context, vpnservice_id)
        driver = self._get_driver_for_vpnservice(context, vpnservice)
        self.service_type_manager.del_resource_associations(
            context, [vpnservice_id])
        driver.delete_vpnservice(context, vpnservice)



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

import neutron.conf.plugins.ml2.drivers.ovn.ovn_conf
import neutron.services.provider_configuration

from neutron_vpnaas.db.vpn import vpn_agentschedulers_db
import neutron_vpnaas.services.vpn.agent
import neutron_vpnaas.services.vpn.device_drivers.ipsec
import neutron_vpnaas.services.vpn.device_drivers.strongswan_ipsec
import neutron_vpnaas.services.vpn.ovn_agent


def list_agent_opts():
    return [
        ('vpnagent',
         neutron_vpnaas.services.vpn.agent.vpn_agent_opts),
        ('ipsec',
         neutron_vpnaas.services.vpn.device_drivers.ipsec.ipsec_opts),
        ('strongswan',
         neutron_vpnaas.services.vpn.device_drivers.strongswan_ipsec.
         strongswan_opts),
        ('pluto',
         neutron_vpnaas.services.vpn.device_drivers.ipsec.pluto_opts)
    ]


def list_ovn_agent_opts():
    return [
        ('vpnagent',
         neutron_vpnaas.services.vpn.ovn_agent.VPN_AGENT_OPTS),
        ('ovs',
         neutron_vpnaas.services.vpn.ovn_agent.OVS_OPTS),
        ('ovn',
         neutron.conf.plugins.ml2.drivers.ovn.ovn_conf.ovn_opts),
        ('ipsec',
         neutron_vpnaas.services.vpn.device_drivers.ipsec.ipsec_opts),
        ('strongswan',
         neutron_vpnaas.services.vpn.device_drivers.strongswan_ipsec.
         strongswan_opts),
        ('pluto',
         neutron_vpnaas.services.vpn.device_drivers.ipsec.pluto_opts)
    ]


def list_opts():
    return [
        ('service_providers',
         neutron.services.provider_configuration.serviceprovider_opts),
        ('',
         vpn_agentschedulers_db.VPN_AGENTS_SCHEDULER_OPTS)
    ]



# Copyright 2023 SysEleven GmbH.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from neutron.api import extensions
from neutron.api import wsgi
from neutron.common.ovn import constants as ovn_constants
from neutron import policy
from neutron.tests.common import helpers
from neutron.tests.common import test_db_base_plugin_v2 as test_plugin
from neutron.tests.unit.api import test_extensions
from neutron.tests.unit.extensions import test_l3
from neutron.tests.unit import testlib_api
from neutron_lib import context
from neutron_lib.db import api as db_api
from neutron_lib import exceptions as n_exc
from neutron_lib.plugins import constants as plugin_constants
from neutron_lib.plugins import directory
from neutron_lib import rpc as n_rpc
from oslo_db import exception as db_exc
import oslo_messaging
from oslo_utils import uuidutils
from sqlalchemy import orm
from webob import exc

from neutron_vpnaas.api.rpc.agentnotifiers import vpn_rpc_agent_api
from neutron_vpnaas.extensions import vpn_agentschedulers
from neutron_vpnaas.services.vpn.common import constants
from neutron_vpnaas.tests.unit.db.vpn import test_vpn_db


VPN_HOSTA = "host-1"
VPN_HOSTB = "host-2"


class VPNAgentSchedulerTestMixIn:
    def _request_list(self, path, admin_context=True,
                      expected_code=exc.HTTPOk.code):
        req = self._path_req(path, admin_context=admin_context)
        res = req.get_response(self.ext_api)
        self.assertEqual(expected_code, res.status_int)
        return self.deserialize(self.fmt, res)

    def _path_req(self, path, method='GET', data=None,
                  query_string=None,
                  admin_context=True):
        content_type = 'application/%s' % self.fmt
        body = None
        if data is not None:  # empty dict is valid
            body = wsgi.Serializer().serialize(data, content_type)
        if admin_context:
            return testlib_api.create_request(
                path, body, content_type, method, query_string=query_string)
        else:
            return testlib_api.create_request(
                path, body, content_type, method, query_string=query_string,
                context=context.Context('', 'tenant_id'))

    def _path_create_request(self, path, data, admin_context=True):
        return self._path_req(path, method='POST', data=data,
                              admin_context=admin_context)

    def _path_show_request(self, path, admin_context=True):
        return self._path_req(path, admin_context=admin_context)

    def _path_delete_request(self, path, admin_context=True):
        return self._path_req(path, method='DELETE',
                              admin_context=admin_context)

    def _path_update_request(self, path, data, admin_context=True):
        return self._path_req(path, method='PUT', data=data,
                              admin_context=admin_context)

    def _list_routers_hosted_by_vpn_agent(self, agent_id,
                                          expected_code=exc.HTTPOk.code,
                                          admin_context=True):
        path = "/agents/{}/{}.{}".format(agent_id,
                                         vpn_agentschedulers.VPN_ROUTERS,
                                         self.fmt)
        return self._request_list(path, expected_code=expected_code,
                                  admin_context=admin_context)

    def _add_router_to_vpn_agent(self, id, router_id,
                                 expected_code=exc.HTTPCreated.code,
                                 admin_context=True):
        path = "/agents/{}/{}.{}".format(id,
                                         vpn_agentschedulers.VPN_ROUTERS,
                                         self.fmt)
        req = self._path_create_request(path,
                                        {'router_id': router_id},
                                        admin_context=admin_context)
        res = req.get_response(self.ext_api)
        self.assertEqual(expected_code, res.status_int)

    def _list_vpn_agents_hosting_router(self, router_id,
                                        expected_code=exc.HTTPOk.code,
                                        admin_context=True):
        path = "/routers/{}/{}.{}".format(router_id,
                                          vpn_agentschedulers.VPN_AGENTS,
                                          self.fmt)
        return self._request_list(path, expected_code=expected_code,
                                  admin_context=admin_context)

    def _remove_router_from_vpn_agent(self, id, router_id,
                                      expected_code=exc.HTTPNoContent.code,
                                      admin_context=True):
        path = "/agents/{}/{}/{}.{}".format(id,
                                            vpn_agentschedulers.VPN_ROUTERS,
                                            router_id,
                                            self.fmt)
        req = self._path_delete_request(path, admin_context=admin_context)
        res = req.get_response(self.ext_api)
        self.assertEqual(expected_code, res.status_int)


class VPNAgentSchedulerTestCaseBase(test_vpn_db.VPNTestMixin,
                                    test_l3.L3NatTestCaseMixin,
                                    VPNAgentSchedulerTestMixIn,
                                    test_plugin.NeutronDbPluginV2TestCase):
    fmt = 'json'

    def setUp(self):
        # NOTE(ivasilevskaya) mocking this way allows some control over mocked
        # client like further method mocking with asserting calls
        self.client_mock = mock.MagicMock(name="mocked client")
        mock.patch.object(
            n_rpc, 'get_client').start().return_value = self.client_mock

        service_plugins = {
            'vpnaas_plugin': 'neutron_vpnaas.services.vpn.ovn_plugin.'
                             'VPNOVNPlugin'}
        plugin_str = 'neutron.tests.unit.extensions.test_l3.TestL3NatIntPlugin'
        super().setUp(plugin_str, service_plugins=service_plugins)

        ext_mgr = extensions.PluginAwareExtensionManager.get_instance()
        self.ext_api = test_extensions.setup_extensions_middleware(ext_mgr)
        self.adminContext = context.get_admin_context()

        self.core_plugin = directory.get_plugin()
        self.core_plugin.get_agents = \
            mock.MagicMock(side_effect=self._get_agents)
        self.core_plugin.get_agent = \
            mock.MagicMock(side_effect=self._get_agent)
        self._agents = {}
        self._vpn_agents_by_host = {}

        self.service_plugin = directory.get_plugin(plugin_constants.VPN)
        policy.init()

    def _get_agents(self, context, filters=None):
        if not filters:
            return self._agents.values()

        agents = []
        for agent in self._agents.values():
            for key, values in filters.items():
                if agent[key] not in values:
                    break
            else:
                agents.append(agent)
        return agents

    def _get_agent(self, context, agent_id):
        try:
            return self._agents[agent_id]
        except KeyError:
            raise n_exc.agent.AgentNotFound(id=agent_id)

    def _get_any_metadata_agent_id(self):
        for agent in self._agents.values():
            if agent['agent_type'] == ovn_constants.OVN_METADATA_AGENT:
                return agent['id']

    def _take_down_vpn_agent(self, host):
        self._vpn_agents_by_host[host]['alive'] = False

    def _get_another_agent_host(self, host):
        for agent in self._vpn_agents_by_host.values():
            if agent['host'] != host:
                return agent['host']

    def _register_agent_states(self):
        self._register_vpn_agent(host=VPN_HOSTA)
        self._register_vpn_agent(host=VPN_HOSTB)
        self._register_metadata_agent(host=VPN_HOSTA)
        self._register_metadata_agent(host=VPN_HOSTB)

    def _register_vpn_agent(self, host=None):
        agent = {
            'id': uuidutils.generate_uuid(),
            'binary': "neutron-ovn-vpn-agent",
            'host': host,
            'availability_zone': helpers.DEFAULT_AZ,
            'topic': 'n/a',
            'configurations': {},
            'start_flag': True,
            'agent_type': constants.AGENT_TYPE_VPN,
            'alive': True,
            'admin_state_up': True}
        self._agents[agent['id']] = agent
        self._vpn_agents_by_host[host] = agent

    def _register_metadata_agent(self, host=None):
        agent = {
            'id': uuidutils.generate_uuid(),
            'binary': "neutron-ovn-metadata-agent",
            'host': host,
            'availability_zone': helpers.DEFAULT_AZ,
            'topic': 'n/a',
            'configurations': {},
            'start_flag': True,
            'agent_type': ovn_constants.OVN_METADATA_AGENT,
            'alive': True,
            'admin_state_up': True}
        self._agents[agent['id']] = agent


class VPNAgentSchedulerTestCase(VPNAgentSchedulerTestCaseBase):
    def _take_down_agent_and_run_reschedule(self, host):
        self._take_down_vpn_agent(host)
        plugin = directory.get_plugin(plugin_constants.VPN)
        plugin.reschedule_vpnservices_from_down_agents()

    def _get_agent_host_by_router(self, router_id):
        agents = self._list_vpn_agents_hosting_router(router_id)
        return agents['agents'][0]['host']

    def test_schedule_router(self):
        self._register_agent_states()
        with self.router() as router:
            router_id = router['router']['id']
            self.service_plugin.schedule_router(self.adminContext, router_id)
            host = self._get_agent_host_by_router(router_id)

        self.assertIn(host, (VPN_HOSTA, VPN_HOSTB))

    def test_router_rescheduler_catches_rpc_db_and_reschedule_exceptions(self):
        self._register_agent_states()
        agent_a_id = self._vpn_agents_by_host[VPN_HOSTA]['id']

        with self.router() as router:
            router_id = router['router']['id']
            self._add_router_to_vpn_agent(agent_a_id, router_id)
            mock.patch.object(
                self.service_plugin, 'reschedule_router',
                side_effect=[
                    db_exc.DBError(), oslo_messaging.RemoteError(),
                    vpn_agentschedulers.RouterReschedulingFailed(
                        router_id='f'),
                    ValueError('this raises'),
                    Exception()
                ]).start()
            self._take_down_agent_and_run_reschedule(VPN_HOSTA)  # DBError
            self._take_down_agent_and_run_reschedule(VPN_HOSTA)  # RemoteError
            self._take_down_agent_and_run_reschedule(VPN_HOSTA)  # schedule err
            self._take_down_agent_and_run_reschedule(VPN_HOSTA)  # Value error
            self._take_down_agent_and_run_reschedule(VPN_HOSTA)  # Exception

    def test_router_rescheduler_catches_exceptions_on_fetching_bindings(self):
        with mock.patch('neutron_lib.context.get_admin_context') as get_ctx:
            mock_ctx = mock.Mock()
            get_ctx.return_value = mock_ctx
            mock_ctx.session.query.side_effect = db_exc.DBError()

            # check that no exception is raised
            self.service_plugin.reschedule_vpnservices_from_down_agents()

    def test_router_rescheduler_iterates_after_reschedule_failure(self):
        self._register_agent_states()
        agent_a = self.service_plugin.get_vpn_agent_on_host(
            self.adminContext, VPN_HOSTA)

        with self.vpnservice() as s1, self.vpnservice() as s2:
            # schedule the services to agent A
            self.service_plugin.auto_schedule_routers(
                self.adminContext, agent_a)

            rs_mock = mock.patch.object(
                self.service_plugin, 'reschedule_router',
                side_effect=vpn_agentschedulers.RouterReschedulingFailed(
                    router_id='f'),
            ).start()
            self._take_down_agent_and_run_reschedule(VPN_HOSTA)
            # make sure both had a reschedule attempt even though first failed
            router_id_1 = s1['vpnservice']['router_id']
            router_id_2 = s2['vpnservice']['router_id']
            rs_mock.assert_has_calls(
                [mock.call(mock.ANY, router_id_1, agent_a),
                 mock.call(mock.ANY, router_id_2, agent_a)],
                any_order=True)

    def test_router_is_not_rescheduled_from_alive_agent(self):
        self._register_agent_states()
        agent_a_id = self._vpn_agents_by_host[VPN_HOSTA]['id']

        with self.router() as router:
            router_id = router['router']['id']
            self._add_router_to_vpn_agent(agent_a_id, router_id)

            patch_func_str = ('neutron_vpnaas.db.vpn.vpn_agentschedulers_db.'
                              'VPNAgentSchedulerDbMixin.reschedule_router')
            with mock.patch(patch_func_str) as rr:
                # take down the unrelated agent and run reschedule check
                self._take_down_agent_and_run_reschedule(VPN_HOSTB)
                self.assertFalse(rr.called)

    def test_router_reschedule_from_dead_agent(self):
        self._register_agent_states()
        agent_a_id = self._vpn_agents_by_host[VPN_HOSTA]['id']

        with self.router() as router:
            router_id = router['router']['id']
            self._add_router_to_vpn_agent(agent_a_id, router_id)
            host_before = self._get_agent_host_by_router(router_id)

            self._take_down_agent_and_run_reschedule(VPN_HOSTA)
            host_after = self._get_agent_host_by_router(router_id)

        self.assertEqual(VPN_HOSTA, host_before)
        self.assertEqual(VPN_HOSTB, host_after)

    def test_router_reschedule_with_write_db_wrap(self):
        self._register_agent_states()
        agent_a = self.service_plugin.get_vpn_agent_on_host(
            self.adminContext, VPN_HOSTA)

        with self.vpnservice() as service:
            # schedule the vpn routers to agent A
            with db_api.CONTEXT_WRITER.using(self.adminContext):
                self.service_plugin.auto_schedule_routers(
                    self.adminContext, agent_a)
            ctxt_mock = mock.MagicMock()
            call_mock = mock.MagicMock(
                side_effect=[oslo_messaging.MessagingTimeout, None])
            ctxt_mock.call = call_mock
            self.client_mock.prepare = mock.MagicMock(return_value=ctxt_mock)
            self._take_down_agent_and_run_reschedule(VPN_HOSTA)
            self.assertEqual(2, call_mock.call_count)
            # make sure vpn service was rescheduled even when first attempt
            # failed to notify VPN agent
            router_id = service['vpnservice']['router_id']
            host = self._get_agent_host_by_router(router_id)

            vpn_agents = self._list_vpn_agents_hosting_router(router_id)
            self.assertEqual(1, len(vpn_agents['agents']))
            self.assertEqual(VPN_HOSTB, host)

    def test_router_reschedule_with_read_db_wrap(self):
        self._register_agent_states()
        agent_a = self.service_plugin.get_vpn_agent_on_host(
            self.adminContext, VPN_HOSTA)

        with self.vpnservice():
            # schedule the vpn routers to agent A
            with db_api.CONTEXT_READER.using(self.adminContext):
                self.assertRaises(
                    TypeError,
                    self.service_plugin.auto_schedule_routers,
                    self.adminContext,
                    agent_a)

    def test_router_reschedule_succeeded_after_failed_notification(self):
        self._register_agent_states()
        agent_a = self.service_plugin.get_vpn_agent_on_host(
            self.adminContext, VPN_HOSTA)

        with self.vpnservice() as service:
            # schedule the vpn routers to agent A
            self.service_plugin.auto_schedule_routers(
                self.adminContext, agent_a)
            ctxt_mock = mock.MagicMock()
            call_mock = mock.MagicMock(
                side_effect=[oslo_messaging.MessagingTimeout, None])
            ctxt_mock.call = call_mock
            self.client_mock.prepare = mock.MagicMock(return_value=ctxt_mock)
            self._take_down_agent_and_run_reschedule(VPN_HOSTA)
            self.assertEqual(2, call_mock.call_count)
            # make sure vpn service was rescheduled even when first attempt
            # failed to notify VPN agent
            router_id = service['vpnservice']['router_id']
            host = self._get_agent_host_by_router(router_id)

            vpn_agents = self._list_vpn_agents_hosting_router(router_id)
            self.assertEqual(1, len(vpn_agents['agents']))
            self.assertEqual(VPN_HOSTB, host)

    def test_router_reschedule_failed_notification_all_attempts(self):
        self._register_agent_states()
        agent_a = self.service_plugin.get_vpn_agent_on_host(
            self.adminContext, VPN_HOSTA)

        with self.vpnservice() as vpnservice:
            # schedule the vpn routers to agent A
            self.service_plugin.auto_schedule_routers(
                self.adminContext, agent_a)
            # mock client.prepare and context.call
            ctxt_mock = mock.MagicMock()
            call_mock = mock.MagicMock(
                side_effect=oslo_messaging.MessagingTimeout)
            ctxt_mock.call = call_mock
            self.client_mock.prepare = mock.MagicMock(return_value=ctxt_mock)
            # perform operations
            self._take_down_agent_and_run_reschedule(VPN_HOSTA)
            self.assertEqual(
                vpn_rpc_agent_api.AGENT_NOTIFY_MAX_ATTEMPTS,
                call_mock.call_count)
            router_id = vpnservice['vpnservice']['router_id']
            vpn_agents = self._list_vpn_agents_hosting_router(router_id)
            self.assertEqual(0, len(vpn_agents['agents']))

    def test_router_auto_schedule_with_hosted(self):
        self._register_agent_states()
        agent_a = self.service_plugin.get_vpn_agent_on_host(
            self.adminContext, VPN_HOSTA)
        agent_b = self.service_plugin.get_vpn_agent_on_host(
            self.adminContext, VPN_HOSTB)

        with self.vpnservice() as vpnservice:
            self._register_agent_states()
            ret_a = self.service_plugin.auto_schedule_routers(
                self.adminContext, agent_a)
            ret_b = self.service_plugin.auto_schedule_routers(
                self.adminContext, agent_b)
            router_id = vpnservice['vpnservice']['router_id']
            vpn_agents = self._list_vpn_agents_hosting_router(router_id)
            host = self._get_agent_host_by_router(router_id)
            self.assertTrue(len(ret_a))
            self.assertIn(router_id, ret_a)
            self.assertFalse(len(ret_b))
        self.assertEqual(1, len(vpn_agents['agents']))
        self.assertEqual(VPN_HOSTA, host)

    def test_add_router_to_vpn_agent(self):
        self._register_agent_states()
        agent_a = self.service_plugin.get_vpn_agent_on_host(
            self.adminContext, VPN_HOSTA)
        agent_a_id = agent_a['id']
        agent_b = self.service_plugin.get_vpn_agent_on_host(
            self.adminContext, VPN_HOSTB)
        agent_b_id = agent_b['id']

        with self.router() as router:
            router_id = router['router']['id']
            num_before_add = len(
                self._list_routers_hosted_by_vpn_agent(
                    agent_a_id)['routers'])
            self._add_router_to_vpn_agent(agent_a_id, router_id)
            # add router again to same agent is fine
            self._add_router_to_vpn_agent(agent_a_id, router_id)
            # add router to a second agent is a conflict
            self._add_router_to_vpn_agent(agent_b_id, router_id,
                                          expected_code=exc.HTTPConflict.code)
            num_after_add = len(
                self._list_routers_hosted_by_vpn_agent(
                    agent_a_id)['routers'])
        self.assertEqual(0, num_before_add)
        self.assertEqual(1, num_after_add)

    def test_add_router_to_vpn_agent_wrong_type(self):
        self._register_agent_states()
        agent_id = self._get_any_metadata_agent_id()

        with self.router() as router:
            router_id = router['router']['id']
            # add_router_to_vpn_agent with a metadata agent id shall fail
            self._add_router_to_vpn_agent(
                agent_id, router_id,
                expected_code=exc.HTTPNotFound.code)

    def _test_add_router_to_vpn_agent_db_error(self, exception):
        self._register_agent_states()
        agent_id = self._vpn_agents_by_host[VPN_HOSTA]['id']

        with self.router() as router, \
                mock.patch.object(orm.Session, 'add', side_effect=exception):
            router_id = router['router']['id']

            self._add_router_to_vpn_agent(
                agent_id, router_id,
                expected_code=exc.HTTPConflict.code)

    def test_add_router_to_vpn_agent_duplicate(self):
        self._test_add_router_to_vpn_agent_db_error(db_exc.DBDuplicateEntry)

    def test_add_router_to_vpn_agent_reference_error(self):
        self._test_add_router_to_vpn_agent_db_error(
            db_exc.DBReferenceError('', '', '', ''))

    def test_add_router_to_vpn_agent_db_error(self):
        self._test_add_router_to_vpn_agent_db_error(db_exc.DBError)

    def test_list_routers_hosted_by_vpn_agent_with_invalid_agent(self):
        invalid_agentid = 'non_existing_agent'
        self._list_routers_hosted_by_vpn_agent(invalid_agentid,
                                               exc.HTTPNotFound.code)

    def test_remove_router_from_vpn_agent(self):
        self._register_agent_states()
        agent_id = self._vpn_agents_by_host[VPN_HOSTA]['id']

        with self.router() as router:
            router_id = router['router']['id']

            self._add_router_to_vpn_agent(agent_id, router_id)
            routers = self._list_routers_hosted_by_vpn_agent(agent_id)
            num_before = len(routers['routers'])

            self._remove_router_from_vpn_agent(agent_id, router_id)
            routers = self._list_routers_hosted_by_vpn_agent(agent_id)
            num_after = len(routers['routers'])

        self.assertEqual(1, num_before)
        self.assertEqual(0, num_after)

    def test_remove_router_from_vpn_agent_wrong_agent(self):
        self._register_agent_states()
        agent_a_id = self._vpn_agents_by_host[VPN_HOSTA]['id']
        agent_b_id = self._vpn_agents_by_host[VPN_HOSTB]['id']

        with self.router() as router:
            router_id = router['router']['id']

            self._add_router_to_vpn_agent(agent_a_id, router_id)
            routers = self._list_routers_hosted_by_vpn_agent(agent_a_id)
            num_before = len(routers['routers'])

            # try to remove router from wrong agent is not an error
            self._remove_router_from_vpn_agent(agent_b_id, router_id)
            routers = self._list_routers_hosted_by_vpn_agent(agent_a_id)
            num_after = len(routers['routers'])

        self.assertEqual(1, num_before)
        self.assertEqual(1, num_after)

    def test_remove_router_from_vpn_agent_unknown_agent(self):
        self._register_agent_states()
        agent_a_id = self._vpn_agents_by_host[VPN_HOSTA]['id']

        with self.router() as router:
            router_id = router['router']['id']

            self._add_router_to_vpn_agent(agent_a_id, router_id)
            routers = self._list_routers_hosted_by_vpn_agent(agent_a_id)
            num_before = len(routers['routers'])

            # try to remove router from unknown agent is an error
            self._remove_router_from_vpn_agent(
                'unknown-agent', router_id,
                expected_code=exc.HTTPNotFound.code)
            routers = self._list_routers_hosted_by_vpn_agent(agent_a_id)
            num_after = len(routers['routers'])

        self.assertEqual(1, num_before)
        self.assertEqual(1, num_after)



# Copyright (c) 2010 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#
# Keystone documentation build configuration file, created by
# sphinx-quickstart on Tue May 18 13:50:15 2010.
#
# This file is execfile()'d with the current directory set to it's containing
# dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import os
import sys

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
BASE_DIR = os.path.dirname(os.path.abspath(__file__))
ROOT_DIR = os.path.abspath(os.path.join(BASE_DIR, "..", ".."))
sys.path.insert(0, ROOT_DIR)

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ['sphinx.ext.autodoc',
              'sphinx.ext.coverage',
              'sphinx.ext.ifconfig',
              'sphinx.ext.graphviz',
              'sphinx.ext.todo',
              'oslo_config.sphinxext',
              'oslo_config.sphinxconfiggen',
              'oslo_policy.sphinxext',
              'oslo_policy.sphinxpolicygen',
              'openstackdocstheme',]

todo_include_todos = True

# Add any paths that contain templates here, relative to this directory.
templates_path = []

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
#source_encoding = 'utf-8'

# The master doctree document.
master_doc = 'index'

# General information about the project.
project = 'Neutron VPNaaS'
copyright = '2011-present, OpenStack Foundation.'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#today = ''
# Else, today_fmt is used as the format for a strftime call.
#today_fmt = '%B %d, %Y'

# List of documents that shouldn't be included in the build.
# unused_docs = []

# List of directories, relative to source directory, that shouldn't be searched
# for source files.
exclude_trees = []

# The reST default role (for this markup: `text`) to use for all documents.
#default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
#add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
show_authors = True

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'native'

# A list of ignored prefixes for module index sorting.
modindex_common_prefix = ['neutron_vpnaas.']

# -- Options for man page output --------------------------------------------

# Grouping the document tree for man pages.
# List of tuples 'sourcefile', 'target', 'title', 'Authors name', 'manual'

#man_pages = [
#    ('man/neutron-server', 'neutron-server', 'Neutron Server',
#     ['OpenStack'], 1)
#]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  Major themes that come with
# Sphinx are currently 'default' and 'sphinxdoc'.
# html_theme_path = ["."]
# html_theme = '_theme'
html_theme = 'openstackdocs'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
#html_theme_path = ['_theme']

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
#html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
#html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
#html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
#html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
#html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
#html_use_modindex = True

# If false, no index is generated.
#html_use_index = True

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
#html_show_sourcelink = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#html_use_opensearch = ''

# If nonempty, this is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = ''

# Output file base name for HTML help builder.
#htmlhelp_basename = 'neutrondoc'


# -- Options for LaTeX output ------------------------------------------------

# The paper size ('letter' or 'a4').
#latex_paper_size = 'letter'

# The font size ('10pt', '11pt' or '12pt').
#latex_font_size = '10pt'

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author,
# documentclass [howto/manual]).
latex_documents = [
    ('index', 'doc-neutron-vpnaas.tex', 'Neutron VPN-as-a-Service Documentation',
     'Neutron development team', 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
#latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
#latex_use_parts = False

# Additional stuff for the LaTeX preamble.
#latex_preamble = ''

# Documents to append as an appendix to all manuals.
#latex_appendices = []

# If false, no module index is generated.
#latex_use_modindex = True

# Disable usage of xindy https://bugzilla.redhat.com/show_bug.cgi?id=1643664
latex_use_xindy = False

latex_domain_indices = False

latex_elements = {
    'extraclassoptions': 'openany,oneside',
    'makeindex': '',
    'printindex': '',
    'preamble': r'\setcounter{tocdepth}{3}',
}

# -- Options for openstackdocstheme -------------------------------------------
openstackdocs_repo_name = 'openstack/neutron-vpnaas'
openstackdocs_pdf_link = True
openstackdocs_auto_name = False
openstackdocs_bug_project = 'neutron'
openstackdocs_bug_tag = 'doc'

# -- Options for oslo_config.sphinxconfiggen ---------------------------------

_config_generator_config_files = [
    'vpn_agent.ini',
    'neutron_vpnaas.conf',
    'ovn_vpn_agent.ini',
]


def _get_config_generator_config_definition(conf):
    config_file_path = '../../etc/oslo-config-generator/%s' % conf
    # oslo_config.sphinxconfiggen appends '.conf.sample' to the filename,
    # strip file extentension (.conf or .ini).
    output_file_path = '_static/config_samples/%s' % conf.rsplit('.', 1)[0]
    return (config_file_path, output_file_path)


config_generator_config_file = [
    _get_config_generator_config_definition(conf)
    for conf in _config_generator_config_files
]

# -- Options for oslo_policy.sphinxpolicygen ---------------------------------

policy_generator_config_file = '../../etc/oslo-policy-generator/policy.conf'
sample_policy_basename = '_static/neutron-vpnaas'



# Copyright 2016 <Yi Jing Zhu/IBM>
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""support local id

Revision ID: 52783a36bd67
Revises: fe637dc3f042
Create Date: 2016-04-26 21:40:40.244196

"""

from alembic import op
import sqlalchemy as sa

from neutron.db import migration


# revision identifiers, used by Alembic.
revision = '52783a36bd67'
down_revision = 'fe637dc3f042'

# milestone identifier, used by neutron-db-manage
neutron_milestone = [migration.NEWTON]


def upgrade():
    op.add_column('ipsec_site_connections',
                  sa.Column('local_id',
                            sa.String(length=255),
                            nullable=True))



#    (c) Copyright 2013 Hewlett-Packard Development Company, L.P.
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.db import constants as db_const
from neutron_lib.db import model_base
import sqlalchemy as sa
from sqlalchemy import orm

from neutron.db.models import l3
from neutron.db import models_v2

from neutron_vpnaas.services.vpn.common import constants


AUTH_ALGORITHM_ENUM_VALUES = [
    'sha1', 'sha256', 'sha384', 'sha512',
    'aes-xcbc', 'aes-cmac',
]

ENCRYPTION_ALGORITHM_ENUM_VALUES = [
    '3des',
    'aes-128', 'aes-192', 'aes-256',
    'aes-128-ctr', 'aes-192-ctr', 'aes-256-ctr',
    'aes-128-ccm-8', 'aes-192-ccm-8', 'aes-256-ccm-8',
    'aes-128-ccm-12', 'aes-192-ccm-12', 'aes-256-ccm-12',
    'aes-128-ccm-16', 'aes-192-ccm-16', 'aes-256-ccm-16',
    'aes-128-gcm-8', 'aes-192-gcm-8', 'aes-256-gcm-8',
    'aes-128-gcm-12', 'aes-192-gcm-12', 'aes-256-gcm-12',
    'aes-128-gcm-16', 'aes-192-gcm-16', 'aes-256-gcm-16',
]

PFS_ENUM_VALUES = [
    'group2', 'group5', 'group14', 'group15',
    'group16', 'group17', 'group18', 'group19', 'group20', 'group21',
    'group22', 'group23', 'group24', 'group25', 'group26', 'group27',
    'group28', 'group29', 'group30', 'group31',
]


class IPsecPeerCidr(model_base.BASEV2):
    """Internal representation of a IPsec Peer Cidrs."""

    cidr = sa.Column(sa.String(32), nullable=False, primary_key=True)
    ipsec_site_connection_id = sa.Column(
        sa.String(36),
        sa.ForeignKey('ipsec_site_connections.id',
                      ondelete="CASCADE"),
        primary_key=True)


class IPsecPolicy(model_base.BASEV2, model_base.HasId, model_base.HasProject):
    """Represents a v2 IPsecPolicy Object."""
    __tablename__ = 'ipsecpolicies'
    name = sa.Column(sa.String(db_const.NAME_FIELD_SIZE))
    description = sa.Column(sa.String(db_const.DESCRIPTION_FIELD_SIZE))
    transform_protocol = sa.Column(sa.Enum("esp", "ah", "ah-esp",
                                           name="ipsec_transform_protocols"),
                                   nullable=False)
    auth_algorithm = sa.Column(sa.Enum(*AUTH_ALGORITHM_ENUM_VALUES,
                                       name="vpn_auth_algorithms"),
                               nullable=False)
    encryption_algorithm = sa.Column(sa.Enum(*ENCRYPTION_ALGORITHM_ENUM_VALUES,
                                             name="vpn_encrypt_algorithms"),
                                     nullable=False)
    encapsulation_mode = sa.Column(sa.Enum("tunnel", "transport",
                                           name="ipsec_encapsulations"),
                                   nullable=False)
    lifetime_units = sa.Column(sa.Enum("seconds", "kilobytes",
                                       name="vpn_lifetime_units"),
                               nullable=False)
    lifetime_value = sa.Column(sa.Integer, nullable=False)
    pfs = sa.Column(sa.Enum(*PFS_ENUM_VALUES, name="vpn_pfs"), nullable=False)


class IKEPolicy(model_base.BASEV2, model_base.HasId, model_base.HasProject):
    """Represents a v2 IKEPolicy Object."""
    __tablename__ = 'ikepolicies'
    name = sa.Column(sa.String(db_const.NAME_FIELD_SIZE))
    description = sa.Column(sa.String(db_const.DESCRIPTION_FIELD_SIZE))
    auth_algorithm = sa.Column(sa.Enum(*AUTH_ALGORITHM_ENUM_VALUES,
                                       name="vpn_auth_algorithms"),
                               nullable=False)
    encryption_algorithm = sa.Column(sa.Enum(*ENCRYPTION_ALGORITHM_ENUM_VALUES,
                                             name="vpn_encrypt_algorithms"),
                                     nullable=False)
    phase1_negotiation_mode = sa.Column(sa.Enum("main", 'aggressive',
                                                name="ike_phase1_mode"),
                                        nullable=False)
    lifetime_units = sa.Column(sa.Enum("seconds", "kilobytes",
                                       name="vpn_lifetime_units"),
                               nullable=False)
    lifetime_value = sa.Column(sa.Integer, nullable=False)
    ike_version = sa.Column(sa.Enum("v1", "v2", name="ike_versions"),
                            nullable=False)
    pfs = sa.Column(sa.Enum(*PFS_ENUM_VALUES, name="vpn_pfs"), nullable=False)


class IPsecSiteConnection(model_base.BASEV2, model_base.HasId,
                          model_base.HasProject):
    """Represents a IPsecSiteConnection Object."""
    __tablename__ = 'ipsec_site_connections'
    name = sa.Column(sa.String(db_const.NAME_FIELD_SIZE))
    description = sa.Column(sa.String(db_const.DESCRIPTION_FIELD_SIZE))
    peer_address = sa.Column(sa.String(255), nullable=False)
    peer_id = sa.Column(sa.String(255), nullable=False)
    local_id = sa.Column(sa.String(255), nullable=True)
    route_mode = sa.Column(sa.String(8), nullable=False)
    mtu = sa.Column(sa.Integer, nullable=False)
    initiator = sa.Column(sa.Enum("bi-directional", "response-only",
                                  name="vpn_initiators"), nullable=False)
    auth_mode = sa.Column(sa.String(16), nullable=False)
    psk = sa.Column(sa.String(255), nullable=False)
    dpd_action = sa.Column(sa.Enum("hold", "clear",
                                   "restart", "disabled",
                                   "restart-by-peer", name="vpn_dpd_actions"),
                           nullable=False)
    dpd_interval = sa.Column(sa.Integer, nullable=False)
    dpd_timeout = sa.Column(sa.Integer, nullable=False)
    status = sa.Column(sa.String(16), nullable=False)
    admin_state_up = sa.Column(sa.Boolean(), nullable=False)
    vpnservice_id = sa.Column(sa.String(36),
                              sa.ForeignKey('vpnservices.id'),
                              nullable=False)
    ipsecpolicy_id = sa.Column(sa.String(36),
                               sa.ForeignKey('ipsecpolicies.id'),
                               nullable=False)
    ikepolicy_id = sa.Column(sa.String(36),
                             sa.ForeignKey('ikepolicies.id'),
                             nullable=False)
    ipsecpolicy = orm.relationship(
        IPsecPolicy, backref='ipsec_site_connection')
    ikepolicy = orm.relationship(IKEPolicy, backref='ipsec_site_connection')
    peer_cidrs = orm.relationship(IPsecPeerCidr,
                                  backref='ipsec_site_connection',
                                  lazy='joined',
                                  cascade='all, delete, delete-orphan')
    local_ep_group_id = sa.Column(sa.String(36),
                                  sa.ForeignKey('vpn_endpoint_groups.id'))
    peer_ep_group_id = sa.Column(sa.String(36),
                                 sa.ForeignKey('vpn_endpoint_groups.id'))
    local_ep_group = orm.relationship("VPNEndpointGroup",
                                      foreign_keys=local_ep_group_id)
    peer_ep_group = orm.relationship("VPNEndpointGroup",
                                     foreign_keys=peer_ep_group_id)


class VPNService(model_base.BASEV2, model_base.HasId, model_base.HasProject):
    """Represents a v2 VPNService Object."""
    name = sa.Column(sa.String(db_const.NAME_FIELD_SIZE))
    description = sa.Column(sa.String(db_const.DESCRIPTION_FIELD_SIZE))
    status = sa.Column(sa.String(16), nullable=False)
    admin_state_up = sa.Column(sa.Boolean(), nullable=False)
    external_v4_ip = sa.Column(sa.String(16))
    external_v6_ip = sa.Column(sa.String(64))
    subnet_id = sa.Column(sa.String(36), sa.ForeignKey('subnets.id'))
    router_id = sa.Column(sa.String(36), sa.ForeignKey('routers.id'),
                          nullable=False)
    subnet = orm.relationship(models_v2.Subnet)
    router = orm.relationship(l3.Router)
    ipsec_site_connections = orm.relationship(
        IPsecSiteConnection,
        backref='vpnservice',
        cascade="all, delete-orphan")
    flavor_id = sa.Column(sa.String(36), sa.ForeignKey(
        'flavors.id', name='fk_vpnservices_flavors_id'))


class VPNEndpoint(model_base.BASEV2):
    """Endpoints used in VPN connections.

    All endpoints in a group must be of the same type. Note: the endpoint
    is an 'opaque' field used to hold different endpoint types, and be
    flexible enough to use for future types.
    """
    __tablename__ = 'vpn_endpoints'
    endpoint = sa.Column(sa.String(255), nullable=False, primary_key=True)
    endpoint_group_id = sa.Column(sa.String(36),
                                  sa.ForeignKey('vpn_endpoint_groups.id',
                                                ondelete="CASCADE"),
                                  primary_key=True)


class VPNEndpointGroup(model_base.BASEV2, model_base.HasId,
                       model_base.HasProject):
    """Collection of endpoints of a specific type, for VPN connections."""
    __tablename__ = 'vpn_endpoint_groups'
    name = sa.Column(sa.String(db_const.NAME_FIELD_SIZE))
    description = sa.Column(sa.String(db_const.DESCRIPTION_FIELD_SIZE))
    endpoint_type = sa.Column(sa.Enum(*constants.VPN_SUPPORTED_ENDPOINT_TYPES,
                                      name="endpoint_type"),
                              nullable=False)
    endpoints = orm.relationship(VPNEndpoint,
                                 backref='endpoint_group',
                                 lazy='joined',
                                 cascade='all, delete, delete-orphan')



# Copyright 2018, Fujitsu Vietnam Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""drop cisco_csr_identifier_map table

Revision ID: e50641731f1a
Revises: b6a2519ab7dc
Create Date: 2018-02-28 10:28:59.846652

"""

from alembic import op
import sqlalchemy as sa

from neutron.db import migration

# revision identifiers, used by Alembic.
revision = 'e50641731f1a'
down_revision = 'b6a2519ab7dc'

# milestone identifier, used by neutron-db-manage
neutron_milestone = [migration.ROCKY, migration.STEIN,
                     migration.TRAIN, migration.USSURI]


def upgrade():
    insp = sa.inspect(op.get_bind())
    if 'cisco_csr_identifier_map' not in insp.get_table_names():
        return
    op.drop_table('cisco_csr_identifier_map')



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from neutron.conf.policies import base as neutron_base
from neutron_lib import policy as base
from oslo_policy import policy

DEPRECATED_REASON = """
The VPaaS API now supports Secure RBAC default roles for ike policies.
"""

rules = [
    policy.DocumentedRuleDefault(
        name='create_ikepolicy',
        check_str=neutron_base.ADMIN_OR_PROJECT_MEMBER,
        scope_types=['project'],
        description='Create an IKE policy',
        operations=[
            {
                'method': 'POST',
                'path': '/vpn/ikepolicies',
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='create_ikepolicy',
            check_str=base.RULE_ANY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since='2025.2')
    ),
    policy.DocumentedRuleDefault(
        name='update_ikepolicy',
        check_str=neutron_base.ADMIN_OR_PROJECT_MEMBER,
        scope_types=['project'],
        description='Update an IKE policy',
        operations=[
            {
                'method': 'PUT',
                'path': '/vpn/ikepolicies/{id}',
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='update_ikepolicy',
            check_str=base.RULE_ADMIN_OR_OWNER,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since='2025.2')
    ),
    policy.DocumentedRuleDefault(
        name='delete_ikepolicy',
        check_str=neutron_base.ADMIN_OR_PROJECT_MEMBER,
        scope_types=['project'],
        description='Delete an IKE policy',
        operations=[
            {
                'method': 'DELETE',
                'path': '/vpn/ikepolicies/{id}',
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='delete_ikepolicy',
            check_str=base.RULE_ADMIN_OR_OWNER,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since='2025.2')
    ),
    policy.DocumentedRuleDefault(
        name='get_ikepolicy',
        check_str=neutron_base.ADMIN_OR_PROJECT_MEMBER,
        scope_types=['project'],
        description='Get IKE policyies',
        operations=[
            {
                'method': 'GET',
                'path': '/vpn/ikepolicies',
            },
            {
                'method': 'GET',
                'path': '/vpn/ikepolicies/{id}',
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='get_ikepolicy',
            check_str=base.RULE_ADMIN_OR_OWNER,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since='2025.2')
    ),
]


def list_rules():
    return rules



# Copyright 2023 SysEleven GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from neutron.plugins.ml2.drivers.ovn.agent import neutron_agent
from neutron.plugins.ml2.drivers.ovn.mech_driver.ovsdb import ovsdb_monitor
from neutron_lib.plugins import constants as plugin_constants
from neutron_lib.plugins import directory

from neutron_vpnaas.services.vpn.common import constants


class OVNVPNAgent(neutron_agent.NeutronAgent):
    agent_type = constants.AGENT_TYPE_VPN
    binary = "neutron-ovn-vpn-agent"

    @property
    def nb_cfg(self):
        return int(self.chassis_private.external_ids.get(
            constants.OVN_AGENT_VPN_SB_CFG_KEY, 0))

    @staticmethod
    def id_from_chassis_private(chassis_private):
        return chassis_private.external_ids.get(
            constants.OVN_AGENT_VPN_ID_KEY)

    @property
    def agent_id(self):
        return self.id_from_chassis_private(self.chassis_private)

    @property
    def description(self):
        return self.chassis_private.external_ids.get(
            constants.OVN_AGENT_VPN_DESC_KEY, '')


class ChassisVPNAgentWriteEvent(ovsdb_monitor.ChassisAgentEvent):
    events = (ovsdb_monitor.BaseEvent.ROW_CREATE,
              ovsdb_monitor.BaseEvent.ROW_UPDATE)

    @staticmethod
    def _vpnagent_nb_cfg(row):
        return int(
            row.external_ids.get(constants.OVN_AGENT_VPN_SB_CFG_KEY, -1))

    @staticmethod
    def agent_id(row):
        return row.external_ids.get(constants.OVN_AGENT_VPN_ID_KEY)

    def match_fn(self, event, row, old=None):
        if not self.agent_id(row):
            # Don't create a cached object with an agent_id of 'None'
            return False
        if event == self.ROW_CREATE:
            return True
        try:
            return self._vpnagent_nb_cfg(row) != self._vpnagent_nb_cfg(old)
        except (AttributeError, KeyError):
            return False

    def run(self, event, row, old):
        neutron_agent.AgentCache().update(constants.AGENT_TYPE_VPN, row,
                                          clear_down=True)


class OVNVPNAgentMonitor:
    def watch_agent_events(self):
        l3_plugin = directory.get_plugin(plugin_constants.L3)
        sb_ovn = l3_plugin._sb_ovn
        if sb_ovn:
            idl = sb_ovn.ovsdb_connection.idl
            if isinstance(idl, ovsdb_monitor.OvnSbIdl):
                idl.notify_handler.watch_event(
                    ChassisVPNAgentWriteEvent(idl.driver))



# Copyright 2017 Red Hat, Inc.
# Copyright 2023 SysEleven GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid

from neutron.agent.linux import external_process
from neutron.common.ovn import utils as ovn_utils
from neutron.conf.plugins.ml2.drivers.ovn import ovn_conf as config
from oslo_log import log as logging
from oslo_service import service
from ovsdbapp.backend.ovs_idl import event as row_event
from ovsdbapp.backend.ovs_idl import vlog

from neutron_vpnaas.agent.ovn.vpn import ovsdb
from neutron_vpnaas.services.vpn.common import constants
from neutron_vpnaas.services.vpn import vpn_service

LOG = logging.getLogger(__name__)

OVN_VPNAGENT_UUID_NAMESPACE = uuid.UUID('e1ce3b12-b1e0-4c81-ba27-07c0fec9c12b')


class ChassisCreateEventBase(row_event.RowEvent):
    """Row create event - Chassis name == our_chassis.

    On connection, we get a dump of all chassis so if we catch a creation
    of our own chassis it has to be a reconnection. In this case, we need
    to do a full sync to make sure that we capture all changes while the
    connection to OVSDB was down.
    """
    table = None

    def __init__(self, vpn_agent):
        self.agent = vpn_agent
        self.first_time = True
        events = (self.ROW_CREATE,)
        super().__init__(
            events, self.table, (('name', '=', self.agent.chassis),))
        self.event_name = self.__class__.__name__

    def run(self, event, row, old):
        if self.first_time:
            self.first_time = False
        else:
            # NOTE(lucasagomes): Re-register the ovn vpn agent
            # with the local chassis in case its entry was re-created
            # (happens when restarting the ovn-controller)
            self.agent.register_vpn_agent()
            LOG.info("Connection to OVSDB established, doing a full sync")
            self.agent.sync()


class ChassisCreateEvent(ChassisCreateEventBase):
    table = 'Chassis'


class ChassisPrivateCreateEvent(ChassisCreateEventBase):
    table = 'Chassis_Private'


class SbGlobalUpdateEvent(row_event.RowEvent):
    """Row update event on SB_Global table."""

    def __init__(self, vpn_agent):
        self.agent = vpn_agent
        table = 'SB_Global'
        events = (self.ROW_UPDATE,)
        super().__init__(events, table, None)
        self.event_name = self.__class__.__name__

    def run(self, event, row, old):
        table = ('Chassis_Private' if self.agent.has_chassis_private
                 else 'Chassis')
        external_ids = {constants.OVN_AGENT_VPN_SB_CFG_KEY: str(row.nb_cfg)}
        self.agent.sb_idl.db_set(
            table, self.agent.chassis,
            ('external_ids', external_ids)).execute()


class OvnVpnAgent(service.Service):
    def __init__(self, conf):
        super().__init__()
        self.conf = conf
        vlog.use_python_logger(max_level=config.get_ovn_ovsdb_log_level())
        self._process_monitor = external_process.ProcessMonitor(
            config=self.conf,
            resource_type='ipsec')

        self.service = vpn_service.VPNService(self)
        self.device_drivers = self.service.load_device_drivers(self.conf.host)

    def _load_config(self):
        self.chassis = self._get_own_chassis_name()
        try:
            self.chassis_id = uuid.UUID(self.chassis)
        except ValueError:
            # OVS system-id could be a non UUID formatted string.
            self.chassis_id = uuid.uuid5(OVN_VPNAGENT_UUID_NAMESPACE,
                                         self.chassis)
        LOG.debug("Loaded chassis name %s (UUID: %s).",
                  self.chassis, self.chassis_id)

    def start(self):
        super().start()

        self.ovs_idl = ovsdb.VPNAgentOvsIdl().start()
        self._load_config()

        tables = ('SB_Global', 'Chassis')
        events = (SbGlobalUpdateEvent(self), )
        # TODO(lucasagomes): Remove this in the future. Try to register
        # the Chassis_Private table, if not present, fallback to the normal
        # Chassis table.
        # Open the connection to OVN SB database.
        self.has_chassis_private = False
        try:
            self.sb_idl = ovsdb.VPNAgentOvnSbIdl(
                chassis=self.chassis, tables=tables + ('Chassis_Private', ),
                events=events + (ChassisPrivateCreateEvent(self), )).start()
            self.has_chassis_private = True
        except AssertionError:
            self.sb_idl = ovsdb.VPNAgentOvnSbIdl(
                chassis=self.chassis, tables=tables,
                events=events + (ChassisCreateEvent(self), )).start()

        # Register the agent with its corresponding Chassis
        self.register_vpn_agent()

        # Do the initial sync.
        self.sync()

    def sync(self):
        for driver in self.device_drivers:
            driver.sync(driver.context, [])

    @ovn_utils.retry()
    def register_vpn_agent(self):
        # NOTE(lucasagomes): db_add() will not overwrite the UUID if
        # it's already set.
        table = ('Chassis_Private' if self.has_chassis_private else 'Chassis')
        # Generate unique, but consistent vpn agent id for chassis name
        agent_id = uuid.uuid5(self.chassis_id, 'vpn_agent')
        ext_ids = {constants.OVN_AGENT_VPN_ID_KEY: str(agent_id)}
        self.sb_idl.db_add(table, self.chassis, 'external_ids',
                           ext_ids).execute(check_error=True)

    def _get_own_chassis_name(self):
        """Return the external_ids:system-id value of the Open_vSwitch table.

        As long as ovn-controller is running on this node, the key is
        guaranteed to exist and will include the chassis name.
        """
        ext_ids = self.ovs_idl.db_get(
            'Open_vSwitch', '.', 'external_ids').execute()
        return ext_ids['system-id']
